/**
 * (1 + exp(A * x + B))^-1 */

#include "sigmoid.h"
#include <stdio.h>
#include <math.h>
#include <vector>
#include <string>
#include <algorithm>
#include <iostream>
#include <fstream>

int split_string(const string &src, const string& key, vector <string>& result)
{
     result.clear();
     int len =  src.size();
     int i = 0; int si = 0;
     while(i < len) {
	while (i < len && key.find(src[i]) != string::npos) { si++; i++; 
	}; // skip beginning spaces
	while (i < len && key.find(src[i]) == string::npos) i++; // skip contents
	    result.push_back(src.substr(si,i-si));
	    si = i;
     }
   
     return result.size();
}

Sigmoid::Sigmoid() {A = 1.0; B = 0.0;}

double Sigmoid::value(double x) 
{
   return 1.0 / (1.0 + exp(A * x + B));
}

void Sigmoid::fitting(double* val_pos, int num_pos, double* val_neg, int num_neg) 
{
  // initialize
  double prior0 = num_pos, prior1 = num_neg;
  double hiTarget = (prior1 + 1.0) / (prior1 + 2.0);
  double loTarget = 1.0 / (prior0 + 2.0);
  double lambda = 1.0e-3;
  double* pp_pos = new double[num_pos];
  double* pp_neg = new double[num_neg];
  double a, b, c, d, e, d1, d2, err, olderr, oldA, oldB;
  double diff, scale, tmp;
  int count = 0, i, j;

  tmp = (prior1 + 1.0) / (prior0 + prior1 + 2.0);
  for (i = num_pos - 1; i >= 0; i--) pp_pos[i] = tmp;
  for (i = num_neg - 1; i >= 0; i--) pp_neg[i] = tmp;
  A = 0.0; B = log((prior0 + 1.0) / (prior1 + 1.0));
  olderr = 1.0e300;

  // main
  for (j = 1; j <= 100; j++) {
    // compute hessian and gradient of error function w.r.t. A and B
    a = b = c = d = e = 0.0;

    // for positive examples
    for (i = num_pos - 1; i >= 0; i--) {
      d1 = pp_pos[i] - hiTarget; d2 = pp_pos[i] * (1.0 - pp_pos[i]);
      a += val_pos[i] * val_pos[i] * d2;
      b += d2;
      c += val_pos[i] * d2;
      d += val_pos[i] * d1;
      e += d1;
    }

    // for negative examples
    for (i = num_neg - 1; i >= 0; i--) {
      d1 = pp_neg[i] - loTarget; d2 = pp_neg[i] * (1.0 - pp_neg[i]);
      a += val_neg[i] * val_neg[i] * d2;
      b += d2;
      c += val_neg[i] * d2;
      d += val_neg[i] * d1;
      e += d1;
    }

    // if gradient is really tiny, then stop
    if (fabs(d) < 1.0e-9 && fabs(e) < 1.0e-9) break;

    oldA = A; oldB = B; err = 0.0;

    // loop until goodness of fit increases
    while (1) {
      double det = (a + lambda) * (b + lambda) - c * c;
      if (det == 0.0) { // if determinant of Hessian is zero
        // increase stabilizer
        lambda *= 10.0;
        continue;
      }

      A = oldA + ((b + lambda) * d - c * e) / det;
      B = oldB + ((a + lambda) * e - c * d) / det;

      // compute the goodness of fit
      err = 0.0;

      // for positive examples
      for (i = num_pos - 1; i >= 0; i--) {
        pp_pos[i] = 1.0 / (1.0 + exp(val_pos[i] * A + B));
        err -= log(pp_pos[i]);
      }

      // for negative examples
      for (i = num_neg - 1; i >= 0; i--) {
        pp_neg[i] = 1.0 / (1.0 + exp(val_neg[i] * A + B));
        err -= log(1.0 - pp_neg[i]);
      }

      if (err < olderr * (1.0 + 1.0e-7)) {
        lambda *= 0.1;
        break;
      }

      // error did not decrease: increase stabilizer and try again
      lambda *= 10.0;
      if (lambda >= 1.0e6) break;
    }

    diff = err - olderr;
    scale = 0.5 * (err + olderr + 1.0);
    if (diff > -1.0e-3 * scale && diff < 1.0e-7 * scale) count++;
    else count = 0;

    olderr = err;
    if (count == 3) break;
  }
}

void Sigmoid::save(FILE* fp) 
{
  fprintf(fp, "%lg %lg\n", A, B);
}

int main(int argc, char **argv) 
{
   if (argc < 1) {
      cerr << "Usage: " << argv[0] << " file " << endl;
      exit(1);
   }
   
   ifstream ifs(argv[argc-1]);
   if (! ifs) {
      cerr << "Cannot open " << argv[argc-1] <<endl;
   }
   
   double *pos = new double [30000];
   double *neg = new double [30000];
   int pos_num = 0;
   int neg_num = 0;
   
   string line;
   vector <string> list;
   vector <string> tmp;
   while(getline(ifs,line)) {
      split_string(line, "\t ", tmp);
      if (tmp.size() != 2) {
	 cerr << "Fomatt Error: " << line << endl;
	 exit(1);
      }
      
      double v1 = atof(tmp[0].c_str());
      double v2 = atof(tmp[1].c_str());

      if (v1 == +1) {
	 pos[pos_num++] = v2;
      } else if (v1 == -1) {
	 neg[neg_num++] = v2;
      } else {
	 cerr << "Fomatt Error: " << line << " ignored" << endl;
      }
   }
   
   Sigmoid sg;
   sg.fitting(pos,pos_num,neg,neg_num);
   sg.save(stdout);
}
